import sys
algo=sys.argv[1]
family=sys.argv[2]
SecurityLevel=int(sys.argv[3])
kappa=int(sys.argv[4])
lgA=float(sys.argv[5])
lgB=float(sys.argv[6])


load taxonomyComplete.sage

"""
given a polynomial h and a polynomial fg (i.e. f or g) and log_2(A), it computes the binary size of the norm of a(t)-x*b(t) where a(t) and b(t) are random polynomials with coefficients less than A and degree less than deg(h) 
"""
def probabilistic_norm_bitsize(fg,h,lgA,N=3*1000):
    Q.<t,x>=QQ['t,x']
    eta=h.degree()
    A=ceil(2^lgA)    
    val=0
    for j in range(N):
        a=sum([randint(-A,A+1)*t^i for i in range(0,eta)])
        b=sum([randint(-A,A+1)*t^i for i in range(0,eta)])
        val+=log((a-b*x).resultant(fg(t,x),x).resultant(h(t,0),t),2).n().abs()
    return val/N


"""
for in the set {z=x+i*jump : z < y}
"""
def frange(x, y, jump):
    while x < y:
        yield x
        x += jump


def polyselect(algo,eta,kappa,P,v,x0,p):
    #print "eta=",eta
    #print "kappa=",kappa    

    # construct f,g
    Q.<t,x>=QQ['t,x']
    if algo == "SexTNFS":
        print "SexTNFS"
        if gcd(eta,kappa) == 1:
            f = P(x=x^kappa)       
            g = x^kappa-v(x=x0)
        else:
            f = P(x=x^kappa+t)
            g = x^kappa+t-v(x=x0)
        if kappa == 2:
            auto = 2
        else:
            auto = 1
    elif algo in ["exTNFS-Conj","exTNFS-GJL"] and kappa == 1:
        print "TNFS base-m"
        lgp = log(p,2).n()
        d = k # floor(sqrt(2*lgp/lgA))-1
        m=floor(p^(1/(d+1)))
        g = x-m
        f = sum([randint(1,m)*x^i for i in range(0,d+1)])
        auto = 1
    elif algo == "exTNFS-Conj":
        print "exTNFS-Conj"
        v = randint(floor(sqrt(p))//2, floor(sqrt(p)))
        u = randint(floor(sqrt(p))//2, floor(sqrt(p)))
        if gcd(eta,kappa) == 1:
            f = x^(2*kappa)+3  # representative
            g = v*(x^kappa) - u
        else:
            f = x^(2*kappa)+3+t
            g = v*(x^kappa+t) - u
        if kappa == 2:
            auto = 2
        else:
            auto = 1
    elif algo == "exTNFS-GJL":
        print "exTNFS-gJL"
        ng = ceil(p^(kappa/(kappa+1)))
        f = x^(kappa+1)+sum([randint(-1,1)*x^i for i in range(kappa+1)])  
        g = sum([randint(-ng,ng)*x^i for i in range(kappa+1)])
        if gcd(eta,kappa) != 1:
            f = f(x=x+t)
            g = g(x=x+t)
        auto = 1

    # construct h 

    if eta == 1:
        h = t
        auto *= 1
        w=1
    elif eta in [6,8,10,12,16,18,20,22,24,28,30,32,36,40,42,46,48,52]:
        ck = min([ck1 for ck1 in range(7,100) if euler_phi(ck1) == eta])
        h=cyclotomic_polynomial(ck)(t)
        auto *= eta
        w=floor(ck/2)
    else:
        h = t^eta - t + 1
        auto *= 1
        w = 1

    return f,g,h,auto,w



"""
main function
given a security level on the curve side, a family, a parameter eta, log_2(A),log_2(B), it selects the polynomial f,g,h and it optimizes log_2(A) and log_2(B) locally around the given values, which must be valid parameters
"""
def precise_tuning(SecurityLevel,eta,lgA,lgB):


    # read family in taxonomy
    Qx.<t,x>=QQ['t,x'] 
    if not family in rs.keys(): 
        print "Unknown family" 
        #raise ValueError
    r = rs[family]
    q = qs[family]
    k = ks[family]       
    P = Ps[family]
    v = vs[family]
    embedding_degree = k
    eta = k // kappa


    # find the smallest numerical example of elliptic curve
    # in the family which has the required security on the curve side 
    rbsize = 2*SecurityLevel
    x0bsize = ceil((rbsize + ceil(log(ZZ(r.denominator()),2)) ) / r.numerator().degree() )
    x0 = 2^x0bsize
    while ( not q(x0) in ZZ) or (not (ZZ(q(x0))).is_prime()):
        x0 += 1 
        if x0 % 10^6 == 0:
            print x0
    p = ZZ(q(x0))
    lgQ = k * log(p,2).n()


    f,g,h,auto,w = polyselect(algo,eta,kappa,P,v,x0,p)
    print f,g,h
    

    # test if (S)exTNFS succeeds with these parameters
    A=ceil(2^lgA)
    B=ceil(2^lgB)
    gNf=probabilistic_norm_bitsize(f,h,lgA)
    gNg=probabilistic_norm_bitsize(g,h,lgA)


    def eval_cost(A,B,probf,probg,auto=None):
        if auto == None:
            auto=eta*kappa/gcd(eta,kappa)
        lgB=log(B,2).n()
        if family == "KSS18" and kappa == 2:
            return 2*B/log(B)*(probf*probg)^(-1)/(kappa) +2^7*B^2/(lgB*kappa)^2/lg(B)^2 
        else:
            return 2*B/log(B)*(probf*probg)^(-1)/(auto) +2^7*B^2/(lgB*auto)^2/log(B)^2

   
    # abort if the default parameters were wrong
    lgNf=probabilistic_norm_bitsize(f,h,lgA)
    lgNg=probabilistic_norm_bitsize(g,h,lgA)
    #print lgNf,lgNg

    min_cost=eval_cost(A,B,dickman_rho(lgNf/lgB).n(),dickman_rho(lgNg/lgB).n(),auto)
    if not ((2*A+1)^(2*eta)/2*dickman_rho(lgNf/lgB).n()*dickman_rho(lgNg/lgB)/w >=  2*B/log(B)):
       print "lgA and lgB are incorrect"
       print log((2*A+1)^(2*eta)/2*dickman_rho(lgNf/lgB).n()*dickman_rho(lgNg/lgB)/w,2).n(), log(2*B/log(B),2).n()
       raise ValueError


    # locally optimize lgA and lgB
    optimal_lgA=lgA
    optimal_lgB=lgB
    optimal_lgNf=lgNf
    optimal_lgNg=lgNg
    for lgA_ in frange(lgA-0.1,lgA+0.1,0.1):
        lgNf=probabilistic_norm_bitsize(f,h,lgA)
        lgNg=probabilistic_norm_bitsize(g,h,lgA)
        #print lgNf,lgNg
        A=ceil(2^lgA_)
        for lgB_ in frange(lgB-1,lgB+1,0.1):
            B=ceil(2^lgB_) 
            cost=eval_cost(A,B,dickman_rho(lgNf/lgB_).n(),dickman_rho(lgNg/lgB_).n(),auto)
            #print log(cost,2).n()
            if (2*A+1)^(2*eta)/2*dickman_rho(lgNf/lgB_).n()*dickman_rho(lgNg/lgB_)/(w) >=  2 * B/log(B) and cost < min_cost:
                min_cost=cost
                print algo+" "+str(family)+" "+str(SecurityLevel)+": "+"new cost",log(min_cost,2).n()
                optimal_lgA=lgA_
                optimal_lgB=lgB_
                optimal_lgNf=lgNf
                optimal_lgNg=lgNg

    return optimal_lgA,optimal_lgB,log(min_cost,2).n(),optimal_lgNf,optimal_lgNg,lgQ


"""
pretty print function
"""
def pad(x):
    s=str(RR(x).n(15))
    s=s+(11-len(s))*" "
    return s



###### MAIN COMPUTATIONS


optimal_lgA,optimal_lgB,s,lgNf,lgNg,lgQ=precise_tuning(SecurityLevel,eta,lgA,lgB)


if algo == "SexTNFS":
    marker = "p"
elif algo == "exTNFS-Conj":
    marker = "c"
elif algo == "exTNFS-GJL":
    marker = "g"
else:
    raise ValueError

gd = open(marker+"-"+family+"-"+str(SecurityLevel)+"-"+str(kappa)+".txt","w")
gd.write(pad(lgQ)+pad(optimal_lgA)+pad(optimal_lgB)+pad(s)+pad(lgNf)+pad(lgNg))

